extern "C"
{
	#include "FroznWar.h"
	#include "fxlib.h"
	#include "stdlib.h"
}

#define __DEBUG(x) { \
		unsigned int key; \
		char ch[20]; \
		SaveDisp(99); \
		sprintf(ch,"%d",x); \
		PopUpWin(1); \
		locate(3,4); \
		Print((unsigned char *)ch); \
		GetKey(&key); \
		RestoreDisp(99); \
		}

#define __DEBUGX(x) { \
		unsigned int key; \
		char ch[20]; \
		SaveDisp(99); \
		sprintf(ch,"%08x",x); \
		PopUpWin(1); \
		locate(3,4); \
		Print((unsigned char *)ch); \
		GetKey(&key); \
		RestoreDisp(99); \
		}

struct BattleStatistics
{
	unsigned killed;
	unsigned dead;

	unsigned exp;
	unsigned money;
};
struct BattleStatistics bstat;

struct Stats
{
	int life;
	int mana;
	int att;
	int def;
};

struct Awards
{
	int exp;
	int money;
};

struct UnitModel
{
	struct Stats stats;
	char *name;
	unsigned char bmp[32];
};

const struct UnitModel models_ally[2] = {
	{ { 10,5,4,1 },"Deblayeur", { 3, 192, 7, 32, 15, 240, 14, 16, 13, 48, 5, 48, 4, 16, 11, 240, 16, 24, 26, 20, 39, 244, 39, 152, 31, 232, 15, 68, 17, 124, 31, 0 } },
	{ { 18,0,6,2 },"Soldat",    { 7, 224, 12, 48, 14, 24, 15, 244, 13, 52, 61, 52, 60, 20, 63, 244, 50, 60, 58, 52, 55, 60, 63, 200, 25, 248, 15, 116, 23, 124, 31, 0 } }
};

const struct UnitModel models_enemy[2] = {
	{ { 12,0,3,0 },"Lapineige", { 0, 0, 2, 0, 5, 64, 5, 160, 4, 32, 6, 160, 6, 160, 12, 96, 23, 208, 28, 72, 4, 40, 14, 48, 9, 48, 15, 144, 0, 240, 0, 0 } },
	{ { 20,0,5,3 },"Goblanc",   { 0, 0, 24, 24, 28, 56, 23, 200, 20, 108, 8, 62, 12, 215, 12, 39, 23, 247, 39, 174, 56, 36, 63, 244, 25, 28, 31, 136, 0, 248, 0, 0 } }
};

const struct Awards models_awards[2] = {
	{ 20, 50 },
	{ 60, 200 }
};

const struct UnitModel *models[4] = {
	NULL,
	models_ally,
	models_enemy,
	NULL
};

class Unit
{
public:
	/*
	Unit sides, ids by side.
		0 NoUnit
		1 Ally
		|	0 Man
		|	1 Soldier
		2 Enemy
		|	0 Rabbit
		|	1 Goblin
		3 Terrain
	*/

	Unit(int side, int id)
		: _side(side), _id(id), _active(false)
	{
		if(!side) return;

		struct Stats s = models[side][id].stats;
		_stats = (struct Stats *)malloc(sizeof(struct Stats));

		_stats->life = s.life;
		_stats->mana = s.mana;
		_stats->att  = s.att;
		_stats->def  = s.def;
	}

	~Unit(void)
	{
		if(_stats) free(_stats);
	}

	bool active(void)
	{
		return _active;
	}

	int id(void)
	{
		return _id;
	}

	int side(void)
	{
		return _side;
	}

	void setActive(bool active)
	{
		_active = active;
	}

	struct Stats *stats(void)
	{
		return _stats;
	}

protected:
	struct Stats *_stats;

private:
	int _side;
	int _id;
	bool _active;
};

class Field
{
public:
	/*
	Select rules.
		0x00 Select_Any
		0x01 Select_Ally
		0x02 Select_Enemy
		0x10 Select_NonEmpty
		0x20 Select_Active
	*/

	enum OnDeathPolicy
	{
		DeleteOnDeath  = 0,
		RestoreOnDeath = 1
	};

	Field(void)
		: policy(DeleteOnDeath)
	{
		unsigned i;
		for(i=0;i<6;i++)
		{
			units[i] = NULL;
			units[i+6] = rand()&1 ? new Unit(2,rand()%2) : NULL;
		}

		x = 0;
		y = 0;
	}

	~Field(void)
	{
		unsigned i;
		for(i=0;i<12;i++) if(units[i]) delete units[i];
	}

	void attack(int main, int target)
	{
		unsigned x1 = 1+17*(main/3) + 59*((main/3)>1), y1 = 1+17*(main%3);
		unsigned x2 = 1+17*(target/3) + 59*((target/3)>1), y2 = 1+17*(target%3);
		unsigned i,j;
		Unit *m = units[main], *t = units[target];
		Stats *ts = t->stats();
		int damage;
		SaveDisp(3);

		Disp_Bitmap(models[m->side()][m->id()].bmp,x1-1,y1,16,16,1);
		Disp_Bitmap(models[m->side()][m->id()].bmp,x1+1,y1,16,16,1);
		Bdisp_PutDisp_DD();
		Sleep(150);
		RestoreDisp(3);
		Bdisp_PutDisp_DD();
		Sleep(150);
		Disp_Bitmap(models[m->side()][m->id()].bmp,x1-1,y1,16,16,1);
		Disp_Bitmap(models[m->side()][m->id()].bmp,x1+1,y1,16,16,1);
		Bdisp_PutDisp_DD();
		Sleep(150);

		RestoreDisp(3);
		for(i=0;i<16;i++) for(j=0;j<16;j++)
		Bdisp_SetPoint_VRAM(x2+j,y2+i,0);
		Bdisp_PutDisp_DD();
		Sleep(200);
		RestoreDisp(3);
		Bdisp_PutDisp_DD();
		Sleep(200);
		for(i=0;i<16;i++) for(j=0;j<16;j++)
		Bdisp_SetPoint_VRAM(x2+j,y2+i,0);
		Bdisp_PutDisp_DD();
		Sleep(200);

		RestoreDisp(3);
		damage = m->stats()->att - ts->def;
		if(damage < 0) damage = 0;
		ts->life -= damage;
		if(ts->life < 0) ts->life = 0;

		if(!ts->life) setUnit(target,NULL);
		m->setActive(false);
	}

	void clear(void)
	{
		unsigned i;
		for(i=0;i<12;i++) setUnit(i,NULL);
	}

	void enemyTurn(void)
	{
		int main, target;

		for(main=6;main<12;main++) if(units[main])
		{
			do target = rand()%6;
			while(!units[target]);
			attack(main,target);
			if(!hasAllies()) return;
		}
	}

	bool hasActiveAllies(void)
	{
		unsigned i;
		for(i=0;i<6;i++) if(units[i] && units[i]->active()) return true;
		return false;
	}

	bool hasAllies(void)
	{
		unsigned i;
		for(i=0;i<6;i++) if(units[i]) return true;
		return false;
	}

	bool hasEnemies(void)
	{
		unsigned i;
		for(i=6;i<12;i++) if(units[i]) return true;
		return false;
	}

	void setAlliesActive(void)
	{
		unsigned i;
		for(i=0;i<6;i++) if(units[i]) units[i]->setActive(true);
	}

	void setDeathPolicy(OnDeathPolicy p_policy)
	{
		policy = p_policy;
	}

	void setUnit(unsigned cell, Unit *unit)
	{
		if(cell > 11) return;
		Unit *u = units[cell];

		if(u)
		{
			if(policy == RestoreOnDeath && u->side() == 1) data.units[u->id()]++;
			else if(policy == DeleteOnDeath)
			{
				if(u->side() == 1) bstat.dead++;
				else if(u->side() == 2)
					bstat.killed++,
					bstat.exp += models_awards[u->id()].exp,
					bstat.money += models_awards[u->id()].money;
			}
			delete u;
			units[cell] = NULL;
		}

		units[cell] = unit;
	}

	void uiDraw(const char *msg = NULL, bool arrow = false)
	{
		unsigned i;
		Unit *u;

		Disp_Clear();

		for(i=1;i<27;i+=2) if(i&1) Disp_Pixel(91,(i<<1)-1,1), Disp_Pixel(91,(i<<1),1);
		for(i=1;i<27;i+=2) if(i&1) Disp_Pixel(36,(i<<1)-1,1), Disp_Pixel(36,(i<<1),1);
		// Bdisp_DrawLineVRAM(36,0,36,53);
		// Bdisp_DrawLineVRAM(91,0,91,53);

		Bdisp_DrawLineVRAM(0,53,127,53);
		if(msg) PrintXY(1,55,(const unsigned char *)msg,0);
		if(hasAllies() && arrow)
			Bdisp_DrawLineVRAM(120,56,120,60),
			Bdisp_DrawLineVRAM(124,58,120,60),
			Bdisp_DrawLineVRAM(124,58,120,56);

		for(i=0;i<12;i++)
		{
			u = units[i];
			if(!u) continue;

			Disp_Bitmap(models[u->side()][u->id()].bmp,1+17*(i/3)+59*((i/3)>1),1+17*(i%3),16,16,1);
		}
	}

	int uiChooseUnit(void)
	{
		unsigned key, i, unit = 0;
		UnitModel m;
		char ch[20];

		for(i=127;i>64;i--)
		{
			Bdisp_DrawLineVRAM(i,0,i,63);
			Bdisp_PutDisp_DD();
			if(i>65) Bdisp_ClearLineVRAM(i,0,i,63);
		}

		Disp_Bitmap(models[1][0].bmp,76,4,16,16,1);
		Disp_Bitmap(models[1][1].bmp,98,4,16,16,1);
		SaveDisp(3);

		while(1)
		{
			m = models[1][unit];
			Stats stats = m.stats;

			RestoreDisp(3);
			Bdisp_AreaReverseVRAM(75+22*unit,3,92+22*unit,20);
			sprintf(ch,"%s (%d)",m.name,data.units[unit]);
			PrintMini(76,24,(const unsigned char *)ch,0);

			PrintMini(76,34,(const unsigned char *)"Vie",0);
			sprintf(ch,"%d",stats.life);
			PrintMini(106,34,(const unsigned char *)ch,0);
			PrintMini(76,41,(const unsigned char *)"Mana",0);
			sprintf(ch,"%d",stats.mana);
			PrintMini(106,41,(const unsigned char *)ch,0);
			PrintMini(76,48,(const unsigned char *)"Att",0);
			sprintf(ch,"%d",stats.att);
			PrintMini(106,48,(const unsigned char *)ch,0);
			PrintMini(76,55,(const unsigned char *)"Def",0);
			sprintf(ch,"%d",stats.def);
			PrintMini(106,55,(const unsigned char *)ch,0);

			while(1)
			{
				GetKey(&key);
				if(key == 30002) return -1;
				if(key == 30004 && data.units[unit]) return unit;
				if(key == 30020 && unit!=0) { unit--; break; }
				if(key == 30021 && unit!=1) { unit++; break; }
			}
		}
	}

	int uiSelectCell(unsigned rule)
	{
		unsigned key, time;
		unsigned tx = x, ty = y;
		Stats *s;
		Unit *u;
		char ch[10];
		int n;

		SaveDisp(3);

		while(1)
		{
			RestoreDisp(3);
			Bdisp_DrawLineVRAM(x,y,x+17,y);
			Bdisp_DrawLineVRAM(x,y,x,y+17);
			Bdisp_DrawLineVRAM(x+17,y,x+17,y+17);
			Bdisp_DrawLineVRAM(x,y+17,x+17,y+17);

			if(x == tx && y == ty)
			{
				n = x/17;
				n -= 3*(n > 1);
				n = 3*n + (y/17);
				u = units[n];

				if(u)
				{
					s = u->stats();

					PrintMini(41,3,(const unsigned char *)models[u->side()][u->id()].name,0);
					PrintMini(41,14,(const unsigned char *)"Vie",0);
					sprintf(ch,"%d",s->life);
					PrintMini(72,14,(const unsigned char *)ch,0);
					PrintMini(41,21,(const unsigned char *)"Mana",0);
					sprintf(ch,"%d",s->mana);
					PrintMini(72,21,(const unsigned char *)ch,0);
					PrintMini(41,28,(const unsigned char *)"Att",0);
					sprintf(ch,"%d",s->att);
					PrintMini(72,28,(const unsigned char *)ch,0);
					PrintMini(41,35,(const unsigned char *)"Def",0);
					sprintf(ch,"%d",s->def);
					PrintMini(72,35,(const unsigned char *)ch,0);
					if(u->active()) PrintMini(41,45,(const unsigned char *)"Actif",0);
				}

				GetKey(&key);
				if(key == 30004 ||key == 30025)
				{
					if((rule & 0x01) && n > 5) continue;
					if((rule & 0x02) && n < 6) continue;
					if((rule & 0x10) && !u) continue;
					if((rule & 0x20) && !u->active()) continue;
					return (key == 30004 ? n : -n-1);
				}
				if(key == 30002) return -102;
				if(key == 30014) return -101;

				if(key == 30020 && x) tx = x-17;
				if(key == 30021 && x != 110) tx = x+17;
				if(key == 30018 && y) ty = y-17;
				if(key == 30023 && y != 34) ty = y+17;
				time = 3;
				if(tx == 34) tx = 93, time = 0;
				if(tx == 76) tx = 17, time = 0;
			}
			else
			{
				Bdisp_PutDisp_DD();
				Sleep(time);
				if(x != tx) x += (x<tx ? +1 : -1);
				if(y != ty) y += (y<ty ? +1 : -1);
			}
		}
	}

private:
	OnDeathPolicy policy;
	Unit *units[12];
	unsigned x, y;
};

void fight(void)
{
	int p,x;
	int main, target;
	Field *field = new Field();

	bstat.killed = 0;
	bstat.dead = 0;
	bstat.exp = 0;
	bstat.money = 0;

	field->setDeathPolicy(Field::RestoreOnDeath);
	while(1)
	{
		field->uiDraw("Choix des unit\xe6\x00as",true);
		tuto(2);
		p = field->uiSelectCell(0x01);

		if(p == -102) continue;
		if(p == -101)
		{
			if(field->hasAllies()) break;
			else continue;
		}

		if(p >= 0)
		{
			x = field->uiChooseUnit();
			if(x != -1)
				field->setUnit(p,new Unit(1,x)),
				data.units[x]--;
		}
		else field->setUnit(-(p+1),NULL);
	}

	field->setDeathPolicy(Field::DeleteOnDeath);
	while(field->hasAllies() && field->hasEnemies())
	{
		field->setAlliesActive();

		do
		{
			do
			{
				field->uiDraw("S\xe6\x00alection unit\xe6\x00a");
				tuto(3);
				main = field->uiSelectCell(0x31);
			} while(main < 0);

			do
			{
				field->uiDraw("S\xe6\x00alection cible");
				target = field->uiSelectCell(0x12);
			} while(target < 0 && target != -102);

			if(target == -102) continue;
			field->attack(main,target);
			if(!field->hasEnemies()) break;
		} while(field->hasActiveAllies());

		if(!field->hasEnemies()) break;
		field->uiDraw("Tour ennemi");
		field->enemyTurn();
	}

	field->setDeathPolicy(Field::RestoreOnDeath);
	field->clear();
	fightend(field->hasAllies());

	delete field;
}

void fightend(int won)
{
	unsigned key;
	char ch[20];

	Disp_Clear();
	PrintXY(13,1,(const unsigned char *)"Bataille termin\346\012e",0);
	Bdisp_AreaReverseVRAM(0,0,127,8);

	PrintXY(5,20,(const unsigned char *)"Morts",0);
	sprintf(ch,"%d",bstat.dead);
	PrintXY(64,20,(const unsigned char *)ch,0);
	PrintXY(5,28,(const unsigned char *)"Tu\xe6\x00as",0);
	sprintf(ch,"%d",bstat.killed);
	PrintXY(64,28,(const unsigned char *)ch,0);
	PrintXY(5,36,(const unsigned char *)"Butin",0);
	sprintf(ch,(const unsigned char *)"%d pi\346\011ces",bstat.money);
	PrintXY(64,36,(const unsigned char *)ch,0);
	PrintXY(5,44,(const unsigned char *)"Exp",0);
	sprintf(ch,(const unsigned char *)"%d",bstat.exp);
	PrintXY(64,44,(const unsigned char *)ch,0);

	PrintXY(98,56,(const unsigned char *)"[EXE]",0);
	GetKey(&key);

	data.money += bstat.money;
	data.exp += bstat.exp;
	if(data.exp > 600) data.exp = 600;
}
